@@ -3,12 +3,13 @@ package Term::Highlight;
 require Exporter;
 
 use strict;
+use sort 'stable';
 
 our @ISA = qw( Exporter );
 
-our @EXPORT = qw( LoadArgs LoadPatterns ClearPatterns Process );
+our @EXPORT = qw( LoadArgs LoadPatterns ClearPatterns GetPatterns Process );
 
-our $VERSION = "1.3";
+our $VERSION = "2.0.0";
 
 
 sub new
@@ -32,16 +33,16 @@ sub ByPositions
 }
 
 
-sub PrintPosition
-{
-    my ( $position, $header ) = @_;
-    #start position, length, start(1) or end(0) of pattern,
-    #pattern number, pattern, fg color id,
-    #bold, bg color id
-    print "$header: $$position[ 0 ], $$position[ 1 ], $$position[ 2 ],
-                    $$position[ 3 ], ${ $$position[ 4 ] }->[ 0 ], ${ $$position[ 4 ] }->[ 1 ],
-                    ${ $$position[ 4 ] }->[ 2 ], ${ $$position[ 4 ] }->[ 3 ]\n";
-}
+#sub PrintPosition
+#{
+    #my ( $position, $header ) = @_;
+    ##start position, length, start(1) or end(0) of pattern,
+    ##pattern number, pattern, fg color id,
+    ##bold, bg color id
+    #print "$header: $$position[ 0 ], $$position[ 1 ], $$position[ 2 ],
+                    #$$position[ 3 ], ${ $$position[ 4 ] }->[ 0 ], ${ $$position[ 4 ] }->[ 1 ],
+                    #${ $$position[ 4 ] }->[ 2 ], ${ $$position[ 4 ] }->[ 3 ]\n";
+#}
 
 
 sub LoadArgs
@@ -53,7 +54,7 @@ sub LoadArgs
     {
         SWITCH_ARGS:
         {
-            if ( $arg =~ /^-(.*)/ )
+            if ( $arg =~ /^-(.+)/ )
             {
                 SWITCH_OPTS:
                 {
@@ -66,13 +67,13 @@ sub LoadArgs
                     if ( $1 eq "ra" )   { $fgcolor = undef; $bold = undef; $bgcolor = undef;
                                           last SWITCH_OPTS }
                     if ( $1 eq "b" )    { $bold = 1; last SWITCH_OPTS }
-                    if ( ( my $orig = $1 ) =~ /\d{1,3}(?=\.1)/ )
+                    if ( ( my $orig = $1 ) =~ /^\d{1,3}(?=\.1$)/ )
                                         { $bgcolor = substr $orig, $-[ 0 ], $+[ 0 ] - $-[ 0 ];
                                           last SWITCH_OPTS }
-                    if ( ( my $orig = $1 ) =~ /\d{1,3}(?=\.0)?/ )
+                    if ( ( my $orig = $1 ) =~ /^\d{1,3}(?=\.0$|$)/ )
                                         { $fgcolor = substr $orig, $-[ 0 ], $+[ 0 ] - $-[ 0 ];
                                           last SWITCH_OPTS }
-                    print __PACKAGE__, " : Warning: unknown option '-$1'\n";
+                    print __PACKAGE__, " : WARNING: unknown option '$arg'\n";
                 }
                 $lastPatternOmitted = 1;
                 last SWITCH_ARGS;
@@ -91,7 +92,7 @@ sub LoadArgs
     }
     #push undefined pattern for last options not followed by any pattern
     #which will apply for the whole string
-    push @{ $self->{ Patterns } }, [ undef, $fgcolor, $bold ] if $lastPatternOmitted;
+    unshift @{ $self->{ Patterns } }, [ undef, $fgcolor, $bold ] if $lastPatternOmitted;
 }
 
 
@@ -109,28 +110,62 @@ sub ClearPatterns
 }
 
 
+sub GetPatterns
+{
+    my $self = shift;
+    $self->{ Patterns };
+}
+
+
 sub FindPositionsOfTags
 {
     my ( $positions, $patterns, $string_ref ) = @_;
     my $result = 0;
     for ( my $i = 0; $i < @$patterns; ++$i )
     {
-        unless ( defined $$patterns[ $i ][ 0 ] )
-        {
-            my $length = length $$string_ref;
-            push @$positions, ( [ 0, $length, 1, $i, \$$patterns[ $i ] ],
-                                [ $length, $length, 0, $i, \$$patterns[ $i ] ] );
-            ++$result;
-        }
-        else
+        if ( defined $$patterns[ $i ][ 0 ] )
         {
             while ( $$string_ref =~ /$$patterns[ $i ][ 0 ]/g )
             {
-                push @$positions, ( [ $-[ 0 ], $+[ 0 ] - $-[ 0 ], 1, $i, \$$patterns[ $i ] ],
-                                    [ $+[ 0 ], $+[ 0 ] - $-[ 0 ], 0, $i, \$$patterns[ $i ] ] );
+                my $length = $+[ 0 ] - $-[ 0 ];
+                next if $length == 0;
+                my $offset = 0;
+                #trailing new lines cause problems: put the color tag before them
+                #BEWARE: in classis Mac OS "\r" was used as the line terminator,
+                #which is not supported here
+                if ( substr( $$string_ref, $+[0] - 1, 1 ) eq "\n" )
+                {
+                    --$length;
+                    ++$offset;
+                    next if $length == 0;
+                    if ( substr( $$string_ref, $+[0] - 2, 1 ) eq "\r" )
+                    {
+                        --$length;
+                        ++$offset;
+                    }
+                }
+                next if $length == 0;
+                push @$positions, ( [ $-[ 0 ], $length, 1, $i, \$$patterns[ $i ] ],
+                                    [ $+[ 0 ] - $offset, $length, 0, $i, \$$patterns[ $i ] ] );
                 ++$result;
             }
         }
+        else
+        {
+            #always mark the line matched in this case, perhaps with no tags
+            ++$result;
+            my $length = length $$string_ref;
+            next if $length == 0;
+            if ( substr( $$string_ref, -1 ) eq "\n" )
+            {
+                --$length;
+                next if $length == 0;
+                --$length if substr( $$string_ref, -2, 1 ) eq "\r";
+            }
+            next if $length == 0;
+            push @$positions, ( [ 0, $length, 1, $i, \$$patterns[ $i ] ],
+                                [ $length, $length, 0, $i, \$$patterns[ $i ] ] );
+        }
     }
     $result;
 }
@@ -194,11 +229,7 @@ sub InsertTags
                 if ( $tagtype eq "term" )
                 {
                     use integer;
-                    # BEWARE: documentation says that escape sequence 2;
-                    # is not supported everywhere. Setting -ra with escape sequence 2;
-                    # makes wrong foreground color on libvte-based terminals whereas letting
-                    # it be undef may not reset bold typeface in some cases
-                    my $bold = ${ $$position[ 4 ] }->[ 2 ] || 2;
+                    my $bold = ${ $$position[ 4 ] }->[ 2 ] || 22;
                     # BEWARE: documentation says that escape sequences 39; and 49;
                     # are not supported everywhere
                     my ( $fgcolor, $bgcolor )= ( "39", "49" );
@@ -211,7 +242,7 @@ sub InsertTags
                                      ${ $$position[ 4 ] }->[ 3 ];
                         $bold = ${ $$position[ 4 ] }->[ 1 ] / 8 if defined
                                      ${ $$position[ 4 ] }->[ 1 ];
-                        $bold ||= ${ $$position[ 4 ] }->[ 2 ] || 2;
+                        $bold ||= ${ $$position[ 4 ] }->[ 2 ] || 22;
                     }
                     #color id in [16..255]
                     else
@@ -267,7 +298,7 @@ sub Process
     #replace all but the last ending tags from @Positions by one-before-the-last starting tag
     #(crucial for terminal color escape sequences algorithm)
     RearrangePositionsOfTags( \@Positions, \@{ $self->{ Patterns } } );
-    
+
     #insert tags into current line
     InsertTags( \@Positions, $String_ref, $$self{ tagtype } );
 
@@ -277,7 +308,7 @@ sub Process
 
 =head1 NAME
 
-Term::Highlight - perl module to highlight regexp patterns on terminals
+Term::Highlight - Perl module to highlight regexp patterns on terminals
 
 =head1 SYNOPSIS
 
@@ -287,9 +318,11 @@ Term::Highlight - perl module to highlight regexp patterns on terminals
 
 =item $obj = Term::Highlight->new( tagtype => $TAGTYPE );
 
+=item $obj->LoadPatterns( \@ptns );
+
 =item $obj->LoadArgs( \@args );
 
-=item $obj->LoadPatterns( \@ptns );
+=item $obj->GetPatterns( );
 
 =item $obj->ClearPatterns( );
 
@@ -299,14 +332,14 @@ Term::Highlight - perl module to highlight regexp patterns on terminals
 
 Currently C<term> and C<term-debug> tagtypes are supported.
 If tagtype is C<term> then boundaries of found patterns will be enclosed in
-terminal color escape sequence tags, if tagtype is C<term-debug> then they
-will be marked by symbolic sequences.
+ANSI terminal color escape sequence tags, if tagtype is C<term-debug> then they
+will be marked by special symbolic sequences.
 
 =head1 DESCRIPTION
 
-Term::Highlight is perl module aimed to support highlighting of patterns
-on color terminals. It supports 256 color terminals and older
-8 color terminals.
+Term::Highlight is a Perl module aimed to support highlighting of regexp
+patterns on color terminals.
+It supports 256 color terminals a well as older 8 color terminals.
 
 =head1 EXPORTS
 
@@ -314,24 +347,30 @@ on color terminals. It supports 256 color terminals and older
 
 =item B<LoadPatterns>
 
-expects reference to array of references to arrays of structure
-[ $pattern, $fg, $bold, $bg ]. Loads patterns to be processed.
-
-=item B<ClearPatterns>
-
-clears loaded patterns.
+expects a reference to an array of references to arrays of type
+[ $pattern, $fg, $bold, $bg ].
+Loads patterns to be processed.
 
 =item B<LoadArgs>
 
-expects array of references to strings. Loads patterns to be processed.
+expects an array of references to strings.
+Loads patterns to be processed.
 This is just a convenient version of C<LoadPatterns>.
-Example of array to be loaded:
-[ "-46", "-25.1", "-i", "\bw.*?\b", "-100" ].
+Example of array to be loaded: [ "-46", "-25.1", "-i", "\bw.*?\b", "-100" ].
+
+=item B<GetPatterns>
+
+returns a reference to the loaded patterns.
+
+=item B<ClearPatterns>
+
+clears the loaded patterns.
 
 =item B<Process>
 
-expects reference to string. Makes substitution of color tags inside the
-string. Returns count of found matches.
+expects a reference to a string.
+Makes substitution of color tags inside the string.
+Returns count of found matches.
 
 =back
 
@@ -345,7 +384,7 @@ A. Radkov, E<lt>alexey.radkov@gmail.comE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2008 by A. Radkov.
+Copyright (C) 2008-2016 by A. Radkov.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.8 or,
